/**
 * Constants
 */
var gc = gc || {};

var UART_9600_bps = 0;
var UART_19200_bps = 1;
var UART_38400_bps = 2;
var UART_57600_bps = 3;
var UART_115200_bps = 4;
var UART_230400_bps = 5;
var UART_300_bps = 6;
var UART_320_bps = 7;
var UART_600_bps = 8;
var UART_1200_bps = 9;
var UART_2400_bps = 10;
var UART_4800_bps = 11;
var UART_460800_bps = 12;
var UART_921600_bps = 13;
var UART_Disable    = 99;

var DEVICE_ADDR = 0x1;


var Cmd_UART_Control = 15; // 0x0F
var Cmd_UART_SetMode = 19; // 0x13
var Cmd_UART_FlexLed_Control = 159;
var Cmd_UART_FlexLed_Read = 160;
var Cmd_UART_FlexLed_Write = 161;
var Cmd_UART_FlexLed_Pattern_Config = 162;
var Cmd_UART_FlexLed_Pattern_Start = 163;
var Cmd_UART_FlexLed_Pattern_Stop = 164;
var Cmd_UART_FlexLed_Watchdog_Control = 165;
var Cmd_UART_FlexLed_Watchdog_Read_Start = 166;
var Cmd_UART_FlexLed_Watchdog_Read_Stop = 167;


var COMMAND_PACKET = 1;
var REPLY_PACKET = 2;
var ERROR_PACKET = 3;
var PAYLOAD_PACKET = 4;
var INTERRUPT_PACKET = 5;

// packet flags
var FLAG_MOREDATA = 1;

var PACKET_IDENTIFIER = 'T'.charCodeAt(0);

var PACKET_ID = 0;
var PACKET_PEC = 1;
var PACKET_PAYLOAD_LEN = 2;
var PACKET_TYPE = 3;
var PACKET_FLAGS = 4;
var PACKET_SEQ_NUM = 5;
var PACKET_STATUS = 6;
var PACKET_COMMAND = 7;
var PACKET_PAYLOAD = 8;

var MAX_PACKET_SIZE = 62; // size, in bytes, of a USB packet
var PACKET_HEADER_SIZE = PACKET_PAYLOAD;
var MAX_PAYLOAD = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;

/**
 * Helper functions
 */
var setDeviceAddress = function (registerModel, buffer, info) {
    var deviceAddrs;
    if (info.deviceAddrs && !isNaN(info.deviceAddrs)) {
        deviceAddrs = +info.deviceAddrs;
        buffer[0] = (deviceAddrs >> 8) & 0xff;
        buffer[1] = deviceAddrs & 0xff;
    }
    else if (registerModel) {
        deviceAddrs = registerModel.getDeviceAddrsForRegister(info);
        buffer[0] = (deviceAddrs >> 8) & 0xff;
        buffer[1] = deviceAddrs & 0xff;
    }
};

var getResult = function (array) {
    var result = 0;
    var canCommDataStart = PACKET_PAYLOAD + 4;
    var endIndex = PACKET_PAYLOAD + (array[PACKET_PAYLOAD_LEN] - 1 || 0); // Omitting CRC - Hence -1.
    var size = endIndex - (gc.tpsglobal.CANConected ? canCommDataStart : PACKET_PAYLOAD);
    var numBits = Math.ceil(size/8) * 8;
    var mask = (1 << numBits) - 1;

    var data = gc.tpsglobal.CANConected ? array.slice(canCommDataStart, endIndex) : array.slice(PACKET_PAYLOAD, endIndex);

    for (var i = gc.tpsglobal.CANConected ? canCommDataStart : PACKET_PAYLOAD; i < endIndex; i++) {
        result = (result << 8) | (array[i] & 0xff);
    }
    return result & mask;
};

var getDeviceAddress = function(addr) {
    var devAddr;
    devAddr = addr & 0xff;
    return devAddr;
}

var parseStringProperty = function(message, value, valueMap)
{
    var stringValue = ("" + value).toLowerCase();
    if (valueMap.hasOwnProperty(stringValue))
    {
        return valueMap[stringValue];
    }
    else
    {
        message = TheValue + value + ForEntry + message + InRegDefs + ' is not supported.  Valid entries are';
        var delimiter = ' "';
        var lastOption;
        for(var option in valueMap)
        {
            if (valueMap.hasOwnProperty(option))
            {
                if (lastOption)
                {
                    message = message + delimiter + lastOption;
                    delimiter = '", "';
                }
                lastOption = option;
            }
        }
        throw message + '", or "' + lastOption + '".';
    }
};


var TheValue = 'The value "';
var ForEntry = '", for entry ';
var InRegDefs = ', in the register-defs.json file';
var parseNumberProperty = function (message, value, min, max) {
    var result = +value;
    if (isNaN(result)) {
        throw TheValue + value + ForEntry + message + InRegDefs + ' is not a number.';
    }
    if (min && result < min) {
        throw TheValue + value + ForEntry + message + InRegDefs + ' must be greater than ' + min + '.';
    }
    if (max && result > max) {
        throw TheValue + value + ForEntry + message + InRegDefs + ' must be less than ' + max + '.';
    }
    return result;
};

var setBytes = function (array, size, value, offset) {
    for (var i = size; i-- > 0;) {
        array[offset + i] = value & 0xff;
        value = value >>> 8;
    }
};



var init = function () {
    /**
     * This code implements an example custom USB2ANY Interface (functionally just a clone of the default I2C interface)
     *
     * @constructor
     * @param settings - The interface settings defined in the system.json
     * @param registerModel - The register model
     * @implements gc.databind.USB2ANY.ICommInterface
     */
    FlexLed_Interface = function (settings, registerModel) {
        this._registerModel = registerModel;
    };

    /**
     * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
     * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
     * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
     */
    gc.databind.USB2ANY.prototype.FLEXLED_Interface = FlexLed_Interface;

    /**
     * Custom interfaces should be extended from the ICommInterface base class
     */
    FlexLed_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

    FlexLed_Interface.prototype._parityMap = { "none": 0, "even": 1, "odd": 2 };
    FlexLed_Interface.prototype._bitDirectionMap = { "lsb": 0, "msb": 1 };

    /**
     * Initialize symbols for device
     *
     * @param settings - The interface settings defined in the system.json
     * @param registerModel - The register model
     */
/*      FlexLed_Interface.prototype.initSymbolsForDevice = function(settings, registerModel) {
         //registerModel.readDeviceAddrsMap('UART', settings);
         var c = gc.databind.PacketCodecFactory.getConstructor(settings.codec);
         if (c && c.prototype && c.prototype.initSymbolsForDevice)
         {
             c.prototype.initSymbolsForDevice(settings, registerModel);
         }
     }; */

    /**
     * Set the communication parameters
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param settings - The interface settings defined in the system.json
     */
    FlexLed_Interface.prototype.control = function (u2a, settings) {
        //this.flexLEDCodec = gc.databind.internal.PacketCodecFactory.create(settings.codec, u2a.sendCommandPacket.bind(u2a, Cmd_UART_FlexLed_Write));

        var baudRate = UART_9600_bps;
        switch (+settings.baudRate) {
            case 19200:
                baudRate = UART_19200_bps;
                break;
            case 38400:
                baudRate = UART_38400_bps;
                break;
            case 57600:
                baudRate = UART_57600_bps;
                break;
            case 115200:
                baudRate = UART_115200_bps;
                break;
            case 230400:
                baudRate = UART_230400_bps;
                break;
            case 300:
                baudRate = UART_300_bps;
                break;
            case 320:
                baudRate = UART_320_bps;
                break;
            case 600:
                baudRate = UART_600_bps;
                break;
            case 1200:
                baudRate = UART_1200_bps;
                break;
            case 2400:
                baudRate = UART_2400_bps;
                break;
            case 4800:
                baudRate = UART_4800_bps;
                break;
            case 460800:
                baudRate = UART_460800_bps;
                break;
            case 921600:
                baudRate = UART_921600_bps;
                break;
        }

        var parity = parseStringProperty('parity', settings.parity, this._parityMap);
        var bitDirection = parseStringProperty('bitDirection', settings.bitDirection, this._bitDirectionMap);
        var characterLength = 8 - parseNumberProperty('characterLength', settings.characterLength || 8, 7, 8);
        var stopBits = parseNumberProperty('stopBits', settings.stopBits || 1, 1, 2) === 1 ? 0 : 1;
        var flexLEDOrCAN = gc.tpsglobal.CANConected ? 0x1 : 0x0;

        // Device Address assignment
        var tpsGlobalComp = document.querySelector("ti-tps-global");
        if(tpsGlobalComp && tpsGlobalComp.deviceAddress){
          DEVICE_ADDR = getDeviceAddress("0x" + tpsGlobalComp.deviceAddress);
        } else {
          DEVICE_ADDR = getDeviceAddress(settings.deviceAddress);
        }
  
        //var u2aSupportComponent = document.querySelector("ti-widget-u2a-version-info"); 

        u2a.sendCommandPacket(Cmd_UART_Control,
            [
                baudRate, parity, bitDirection, characterLength, stopBits
            ]);


        u2a.sendCommandPacket(Cmd_UART_SetMode, [0,0]);

        u2a.sendCommandPacket(Cmd_UART_FlexLed_Control, [0x55, DEVICE_ADDR, 0x1, 0x31, flexLEDOrCAN]); // TODO : Add Device Address sweep Do Device address sweep and do initialize, Option to change the device address.
        u2a.sendCommandPacket(Cmd_UART_FlexLed_Pattern_Config, [12, 0x00, 0x20, 0x40]);
        
        /* if(u2a.version !== u2aSupportComponent.supportedFirmware){
            u2aSupportComponent.connectionStatusChanged(true);
        } */

        //Reading CONF_CRC register and checking whether device is reponding with value > 0, else disconnect
        return this.read(u2a, {addr:0x78}).then(readVal => {
            if(readVal === 0x55){
                return this.read(u2a, {addr:0x50}).then(enableValue => {
                    if(enableValue === 0x55){
                        console.log("Device Disconnected when reading CRC register and found it's CAN Connection");
                        document.querySelector("#global_pg_toast_error").message = "Please check the CAN checkbox and try connecting again.";
                        document.querySelector("#global_pg_toast_error").show = true;
                        throw "Please check the CAN checkbox and try connecting again.";
                    }
                });
            } else {
                if(readVal === 0){
                    document.querySelector("#global_pg_toast_error").message = "Please check the Device Address (or) correct the CAN Mode status and try connecting again.";
                    document.querySelector("#global_pg_toast_error").show = true;
                    throw "Please check the Device Address (or) correct the CAN Mode status and try connecting again.";
                }
            } 
        });

        /* this.read(u2a, {addr:0x78}).then(readVal => {
            if(readVal === 0x55){
                this.read(u2a, {addr:0x50}).then(enableValue => {
                    if(enableValue === 0x55){
                        gc.connectionManager.disconnect().then(()=>{
                            console.log("Device Disconnected when reading CRC register and found it's CAN Connection");
                            document.querySelector("#global_pg_toast_error").message = "Please check the CAN checkbox and try connecting again.";
                            document.querySelector("#global_pg_toast_error").show = true;
                            throw "Please check the CAN checkbox and try connecting again.";
                        }).fail((e) => {
                           console.log(e);
                       });
                    } else {
                        document.querySelector("ti-tps-global").checkDeviceSettings = true;
                        document.querySelector("#global_pg_toast_error").show = false;
                        document.querySelector("ti-tps-global").isEvmConnected = true;
                    }
                }).catch(error => console.log(error));
            } else {
                if(readVal === 0){
                    gc.connectionManager.disconnect().then(()=>{
                        console.log("Device Disconnected when reading CRC register and found it's 0");
                        document.querySelector("#global_pg_toast_error").message = "Please check the Device Address (or) correct the CAN Mode status and try connecting again.";
                        document.querySelector("#global_pg_toast_error").show = true;
                        throw "Please check the Device Address (or) correct the CAN Mode status and try connecting again.";
                    }).fail((e) => {
                       console.log(e);
                   });
                } else {
                    document.querySelector("ti-tps-global").checkDeviceSettings = true;
                    document.querySelector("#global_pg_toast_error").show = false;
                    document.querySelector("ti-tps-global").isEvmConnected = true;
                }
            } 
        }).catch(error => console.log(error)); */

        //return this.flexLEDCodec.connect(settings);
    };

    /**
     * Construct and send a command packet for reading from a register
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param info - The register map info for the register
     * @return - A promise that is resolved once the data is received
     */
    FlexLed_Interface.prototype.read = function (u2a, info) {

        var readData = [];
        readData[0] = 0;
        readData[1] = info.addr;
        readData[2] = gc.tpsglobal.CANConected ? 6 : 2; // RX data length to be expected
 
        return u2a.readResponse(u2a.sendCommandPacket(Cmd_UART_FlexLed_Read, readData)).then(getResult);
    };

    /**
     * Construct and send a command packet for writing to a register
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param info - The register map info for the register
     * @param value - The value to write to the register
     * @return - A promise that resolves immediately
     */
    FlexLed_Interface.prototype.write = function (u2a, info, value) {
        var writeData = [];
        writeData[0] = 0;
        writeData[1] = (info.writeAddr === undefined ? info.addr : info.writeAddr) || 0;
        writeData[2] = gc.tpsglobal.CANConected ? 7 : 2; //RX Data Length to be expected
        writeData[3] = value;

        u2a.sendCommandPacket(Cmd_UART_FlexLed_Write, writeData);
        return Q(true);
    };
};


// Wait until USB2ANY global var is ready before defining interface
if (gc.databind.USB2ANY) {
    init();
} else {
    document.addEventListener('DOMContentLoaded', init);
}